#!/usr/bin/env python3

# Copyright 2024 Espressif Systems (Shanghai) PTE LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
contains utility functions for generating csr, build certificates, certs/keys conversion etc.
"""
import logging
from typing import Optional
from cryptography import x509
from datetime import datetime, timedelta, timezone
from cryptography.x509.oid import NameOID
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.x509 import load_pem_x509_certificate, load_der_x509_certificate
from cryptography.exceptions import InvalidSignature
from utils import VERIFY_OR_EXIT, VERIFY_OR_RAISE

# CHIP OID for vendor id
VENDOR_ID_OID = x509.ObjectIdentifier('1.3.6.1.4.1.37244.2.1')

# CHIP OID for product id
PRODUCT_ID_OID = x509.ObjectIdentifier('1.3.6.1.4.1.37244.2.2')

VALID_DAYS = 365 * 100

CERTIFICATE_VERSION = x509.Version.v3
SIGNATURE_ALGORITHM = "ecdsa-with-SHA256"

def save_to_file(file_name, file_data):
    """
    Save filedata to file
    Args:
    file_name (str): file name for the file.
    file_data (bytes): file data to store in file.
    """
    with open(file_name, 'wb') as file:
        file.write(file_data)

def chip_gen_ec_key():
    """
    Generate EC private key.
    Returns:
        ec.EllipticCurvePrivateKey: Generated private key object.
    """
    return ec.generate_private_key(ec.SECP256R1())

def load_cert_from_file(cert_file):
    """
    Load an X.509 certificate from file in PEM or DER format.
    Args:
        cert_file (str): Path to the certificate file.
    Returns:
        x509.Certificate: Parsed certificate object in PEM or DER format.
    Raises:
        ValueError: If the certificate format is invalid or unsupported.
    """
    with open(cert_file, 'rb') as file:
        cert_data = file.read()

    # Attempt to load as PEM first, then fallback to DER
    for loader in (load_pem_x509_certificate, load_der_x509_certificate):
        try:
            return loader(cert_data)
        except ValueError:
            continue

    raise ValueError("Invalid certificate format. Supported formats: PEM, DER.")

def load_key_from_file(key_file):
    """
    Load a private key from file in PEM or DER format.
    Args:
        key_file (str): Path to the private key file.
    Returns:
        ec.EllipticCurvePrivateKey: Parsed private key object in PEM or DER format.
    Raises:
        ValueError: If the key format is invalid or unsupported.
    """
    with open(key_file, 'rb') as file:
        key_data = file.read()

    # Attempt to load as PEM first, then fallback to DER
    for loader in (serialization.load_pem_private_key, serialization.load_der_private_key):
        try:
            return loader(key_data, password=None)
        except ValueError:
            continue

    raise ValueError("Invalid key format or unsupported key type. Supported formats: PEM, DER.")

def convert_x509_cert_from_pem_to_der(pem_file, out_der_file):
    """
    Convert an X509 certificate from PEM to DER format.
    Args:
        pem_file (str): Path to the PEM file.
        out_der_file (str): Output path for the DER file.
    """
    cert = load_cert_from_file(pem_file)
    der_cert = cert.public_bytes(serialization.Encoding.DER)
    save_to_file(out_der_file, der_cert)

def convert_private_key_from_pem_to_der(pem_file, out_der_file):
    """
    Convert a private key from PEM to DER format.
    Args:
        pem_file (str): Path to the PEM file.
        out_der_file (str): Output path for the DER file.
    """
    key = load_key_from_file(pem_file)
    der_key = key.private_bytes(
        encoding=serialization.Encoding.DER,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption(),
    )
    save_to_file(out_der_file, der_key)

def store_keypair_as_raw(pem_file, out_privkey_bin, out_pubkey_bin):
    """
    Generate a binary format key pair from a PEM file.
    Args:
        pem_file (str): Path to the PEM file containing the private key.
        out_privkey_bin (str): Output path for the private key in binary format.
        out_pubkey_bin (str): Output path for the public key in binary format.
    """
    key = load_key_from_file(pem_file)
    priv_val = key.private_numbers().private_value
    pub_numbers = key.public_key().public_numbers()

    priv_key_data = priv_val.to_bytes(32, byteorder='big')
    save_to_file(out_privkey_bin, priv_key_data)

    with open(out_pubkey_bin, 'wb') as pub_file:
        pub_file.write(b'\x04')
        pub_file.write(pub_numbers.x.to_bytes(32, byteorder='big'))
        pub_file.write(pub_numbers.y.to_bytes(32, byteorder='big'))

def extract_matter_rdn(subject, oid):
    """
    Extract the value of a custom OID from a certificate.
    Args:
        subject (x509.Name): Subject from which to extract the OID value.
        oid (x509.ObjectIdentifier): OID to extract.
    Returns:
        str: Extracted OID value or None if not found.
    """
    try:
        return subject.get_attributes_for_oid(oid)[0].value
    except IndexError:
        return None

def extract_pid(subject):
    """
    Extract the Product ID (PID) from a certificate's subject.
    Args:
        subject (x509.Name): Subject from which to extract the PID.
    Returns:
        str: Extracted PID or None if not found.
    """
    return extract_matter_rdn(subject, PRODUCT_ID_OID)

def extract_vid(subject):
    """
    Extract the Vendor ID (VID) from a certificate's subject.
    Args:
        subject (x509.Name): Subject from which to extract the VID.
    Returns:
        str: Extracted VID or None if not found.
    """
    return extract_matter_rdn(subject, VENDOR_ID_OID)

def extract_common_name(subject) -> Optional[str]:
    """
    Extract the Common Name (CN) from a certificate's subject.
    Args:
        subject (x509.Name): Subject from which to extract the CN.
    Returns:
        str: Extracted CN or None if not found.
    """
    return extract_matter_rdn(subject, x509.NameOID.COMMON_NAME)

def generate_cert_validity(valid_from, lifetime):
    """
    Generate a certificate validity period.
    Args:
        valid_from (str): Start date for the certificate validity period in ISO 8601 format ("YYYY-MM-DDTHH:MM:SS").
        lifetime (int): Lifetime of the certificate in days.
    Returns:
        tuple: Tuple containing the start and end dates of the certificate validity period as UTC-aware datetime objects.
    """
    nvb_time = datetime.now(timezone.utc)
    nva_time = nvb_time + timedelta(days=VALID_DAYS)
    if valid_from:
        nvb_time = datetime.strptime(str(valid_from), "%Y-%m-%dT%H:%M:%S").replace(tzinfo=timezone.utc)
    if lifetime:
        nva_time = nvb_time + timedelta(days=lifetime)
    return nvb_time, nva_time

def generate_cert_subject(vendor_id, product_id, common_name):
    """
    Generate a certificate subject.
    Args:
        vendor_id (str): Vendor ID
        product_id (str): Product ID
        common_name (str): Common Name
    Returns:
        x509.Name: Generated certificate subject.
    """
    x509_attrs = []
    x509_attrs.append(x509.NameAttribute(NameOID.COMMON_NAME, common_name))
    x509_attrs.append(x509.NameAttribute(VENDOR_ID_OID, vendor_id.zfill(4)))
    x509_attrs.append(x509.NameAttribute(PRODUCT_ID_OID, product_id.zfill(4)))
    return x509.Name(x509_attrs)

def build_certificate(vendor_id:str, product_id:str, ca_cert_file: str, ca_privkey_file: str, out_cert_file: str, out_key_file: str, is_pai: bool, common_name: str, valid_from=None, lifetime=None):
    """
    Build a certificate for a Matter device.

    Args:
        vendor_id (str): vendor id
        product_id (str): product id
        ca_cert_file (str): Path to the CA certificate file used for signing.
        ca_privkey_file (str): Path to the CA private key file used for signing.
        out_cert_file (str): Path to save the generated certificate.
        out_key_file (str): Path to save the private key for the certificate.
        is_pai (bool): Indicates if the certificate is a Product Attestation Intermediate (PAI) certificate.
        common_name (str): Common Name for the certificate subject.
        valid_from (str, optional): Start date for the certificate validity period in ISO 8601 format ("YYYY-MM-DDTHH:MM:SS").
            Defaults to None, in which case the current date is used.
        lifetime (int, optional): Lifetime of the certificate in days. Defaults to None, in which case a default period (e.g., 100 years) is used.

    Returns:
        None: Saves the generated certificate and key to the specified files.
    """
    if not ca_cert_file or not ca_privkey_file:
        raise ValueError("CA key and certificate cannot be None.")

    try:
        # Load CA certificate and key
        ca_cert = load_cert_from_file(ca_cert_file)
        ca_key = load_key_from_file(ca_privkey_file)

        # Generate a new private key for the device certificate
        private_key = chip_gen_ec_key()
        public_key = private_key.public_key()

        # Build subject attributes
        cert_subject = generate_cert_subject(vendor_id, product_id, common_name)

        nvb_time, nva_time = generate_cert_validity(valid_from, lifetime)

        # Build certificate
        cert = x509.CertificateBuilder()
        cert = cert.subject_name(cert_subject)
        cert = cert.issuer_name(ca_cert.subject)
        cert = cert.public_key(public_key)
        cert = cert.serial_number(x509.random_serial_number())
        cert = cert.not_valid_before(nvb_time)
        cert = cert.not_valid_after(nva_time)

        # Extensions for PAI certificate
        if is_pai:
            cert = cert.add_extension(x509.BasicConstraints(ca=True, path_length=0), critical=True)
            cert = cert.add_extension(x509.KeyUsage(digital_signature=True, content_commitment=False,
                                                    key_encipherment=False, data_encipherment=False,
                                                    key_agreement=False, key_cert_sign=True, crl_sign=True,
                                                    encipher_only=False, decipher_only=False), critical=True)
        # Extensions for DAC certificate
        else:
            cert = cert.add_extension(x509.BasicConstraints(ca=False, path_length=None), critical=True)
            cert = cert.add_extension(x509.KeyUsage(digital_signature=True, content_commitment=False,
                                                    key_encipherment=False, data_encipherment=False,
                                                    key_agreement=False, key_cert_sign=False, crl_sign=False,
                                                    encipher_only=False, decipher_only=False), critical=True)

        cert = cert.add_extension(x509.SubjectKeyIdentifier.from_public_key(public_key), critical=False)
        cert = cert.add_extension(x509.AuthorityKeyIdentifier.from_issuer_public_key(ca_key.public_key()), critical=False)

        # Sign the certificate
        signed_cert = cert.sign(private_key=ca_key, algorithm=hashes.SHA256(), backend=default_backend())

        # Save the certificate to a file
        certificate_pem = signed_cert.public_bytes(serialization.Encoding.PEM)
        save_to_file(out_cert_file, certificate_pem)

        # Save the private key to a file
        private_key_pem = private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption()
        )
        save_to_file(out_key_file, private_key_pem)

    except Exception as e:
        raise RuntimeError(f"Failed to build certificate: {e}")

def validate_certificate_validity(valid_from: str, lifetime: int, cert_file: str) -> bool:
    """
    Validate the certificate validity.
    Args:
        valid_from (str): Start date for the certificate validity period in ISO 8601 format ("YYYY-MM-DDTHH:MM:SS").
        lifetime (int): Lifetime of the certificate in days.
        cert_file (str): Path to the certificate file.
    Returns:
        bool: True if the certificate validity is valid, False otherwise.
    """
    try:
        cert = load_cert_from_file(cert_file)
        nvb_time, nva_time = generate_cert_validity(valid_from, lifetime)
        VERIFY_OR_RAISE(cert.not_valid_before_utc <= nvb_time <= cert.not_valid_after_utc,
                       f"Specified start date ({nvb_time}) is outside certificate's validity period ({cert.not_valid_before_utc} to {cert.not_valid_after_utc})")
        VERIFY_OR_RAISE(cert.not_valid_before_utc <= nva_time <= cert.not_valid_after_utc,
                        f"Specified end date based on lifetime ({nva_time}) is outside certificate's validity period ({cert.not_valid_before_utc} to {cert.not_valid_after_utc})")
        return True
    except Exception as e:
        logging.error(f"Certificate validity validation failed for {cert_file}: {str(e)}")
        return False

def validate_vid_pid_with_pai_cert(vid: int, pid: int, pai_cert_file: str) -> bool:
    """
    Validate input VID and PID with the PAI certificate.
    Args:
        vid (int): Input Vendor ID
        pid (int): Input Product ID
        pai_cert_file (str): Input PAI certificate file
    Returns:
        bool: True if the VID and PID match the VID and PID in the PAI certificate, False otherwise.
    Raises:
        ValueError: If the certificate is missing required VID/PID or values don't match.
    """
    try:
        cert = load_cert_from_file(pai_cert_file)
        extracted_vid = extract_vid(cert.subject)
        extracted_pid = extract_pid(cert.subject)

        VERIFY_OR_RAISE(extracted_vid, "Missing VID in PAI certificate")
        try:
            extracted_vid_int = int(extracted_vid, 16)
            VERIFY_OR_RAISE(extracted_vid_int == vid,
                           f"VID: {extracted_vid_int} (0x{extracted_vid}) in PAI certificate does not match input VID: {vid} (0x{vid:04x})")
        except ValueError:
            VERIFY_OR_RAISE(False, f"Invalid VID format in certificate: {extracted_vid}")

        if extracted_pid is not None:
            try:
                extracted_pid_int = int(extracted_pid, 16)
                VERIFY_OR_RAISE(extracted_pid_int == pid,
                               f"PID: {extracted_pid_int} (0x{extracted_pid}) in PAI certificate does not match input PID: {pid} (0x{pid:04x})")
            except ValueError:
                VERIFY_OR_RAISE(False, f"Invalid PID format in certificate: {extracted_pid}")
        return True
    except Exception as e:
        logging.error(f"VID/PID validation failed: {str(e)}")
        return False

def validate_dac_cert(dac_cert_file: str) -> bool:
    """
    Validate the DAC certificate.
    Args:
        dac_cert_file (str): Path to the DAC certificate file.
    Returns:
        bool: True if the DAC certificate is valid, False otherwise.
    """
    try:
        dac_cert = load_cert_from_file(dac_cert_file)

        VERIFY_OR_RAISE(dac_cert.version == CERTIFICATE_VERSION,
                       f"DAC certificate version must be {CERTIFICATE_VERSION}, got {dac_cert.version}")
        VERIFY_OR_RAISE(dac_cert.signature_algorithm_oid._name == SIGNATURE_ALGORITHM,
                       f"DAC certificate signature algorithm must be {SIGNATURE_ALGORITHM}, got {dac_cert.signature_algorithm_oid._name}")

        vendor_id = extract_vid(dac_cert.subject)
        product_id = extract_pid(dac_cert.subject)
        VERIFY_OR_RAISE(vendor_id and product_id, "Missing VID or PID in DAC certificate")


        issuer_vid = extract_vid(dac_cert.issuer)
        issuer_pid = extract_pid(dac_cert.issuer)
        VERIFY_OR_RAISE(issuer_vid, "Missing VID in DAC certificate issuer")
        VERIFY_OR_RAISE(issuer_vid == vendor_id,
                       f"DAC certificate issuer's VID '{issuer_vid}' must match the certificate's VID '{vendor_id}'")
        if issuer_pid is not None:
            VERIFY_OR_RAISE(issuer_pid == product_id,
                           f"DAC certificate issuer's PID '{issuer_pid}' must match the certificate's PID '{product_id}'")

        subject_public_key_info = dac_cert.public_key().curve.name
        VERIFY_OR_RAISE(subject_public_key_info == "secp256r1",
                       f"The algorithm in subjectPublicKeyInfo field must be prime256v1, got {subject_public_key_info}")

        try:
            basic_constraints = dac_cert.extensions.get_extension_for_oid(x509.ExtensionOID.BASIC_CONSTRAINTS)
            VERIFY_OR_RAISE(basic_constraints.critical, "DAC Basic Constraints extension must be marked as critical")
            VERIFY_OR_RAISE(not basic_constraints.value.ca, "DAC must not be a CA certificate")

            key_usage = dac_cert.extensions.get_extension_for_oid(x509.ExtensionOID.KEY_USAGE)
            VERIFY_OR_RAISE(key_usage.critical, "DAC Key Usage extension must be marked as critical")
            VERIFY_OR_RAISE(key_usage.value.digital_signature, "DAC must have digital signature usage enabled")
            VERIFY_OR_RAISE(not (key_usage.value.content_commitment or
                                 key_usage.value.key_encipherment or
                                 key_usage.value.data_encipherment or
                                 key_usage.value.key_agreement or
                                 key_usage.value.crl_sign or
                                 key_usage.value.key_cert_sign), "DAC must not have unsupported key usage extensions")

            subject_key_id = dac_cert.extensions.get_extension_for_oid(x509.ExtensionOID.SUBJECT_KEY_IDENTIFIER)
            authority_key_id = dac_cert.extensions.get_extension_for_oid(x509.ExtensionOID.AUTHORITY_KEY_IDENTIFIER)
            VERIFY_OR_RAISE(subject_key_id, "DAC missing Subject Key Identifier extension")
            VERIFY_OR_RAISE(authority_key_id, "DAC missing Authority Key Identifier extension")
        except x509.ExtensionNotFound as e:
            logging.error(f"DAC missing required extension: {str(e)}")
            return False
        return True

    except Exception as e:
        logging.error(f"DAC certificate validation failed: {str(e)}")
        return False

def validate_pai_cert(pai_cert_file: str) -> bool:
    """
    Validate the PAI certificate.
    Args:
        pai_cert_file (str): Path to the PAI certificate file.
    Returns:
        bool: True if the PAI certificate is valid, False otherwise.
    """
    try:
        pai_cert = load_cert_from_file(pai_cert_file)

        VERIFY_OR_RAISE(pai_cert.version == CERTIFICATE_VERSION,
                       f"PAI certificate version must be {CERTIFICATE_VERSION}, got {pai_cert.version}")
        VERIFY_OR_RAISE(pai_cert.signature_algorithm_oid._name == SIGNATURE_ALGORITHM,
                       f"PAI certificate signature algorithm must be {SIGNATURE_ALGORITHM}, got {pai_cert.signature_algorithm_oid._name}")

        vendor_id = extract_vid(pai_cert.subject)
        VERIFY_OR_RAISE(vendor_id, "Missing VID in PAI certificate")

        issuer_vid = extract_vid(pai_cert.issuer)
        if issuer_vid is not None:
            VERIFY_OR_RAISE(issuer_vid == vendor_id,
                           f"PAI certificate issuer's VID '{issuer_vid}' must match the certificate's VID '{vendor_id}'")

        subject_public_key_info = pai_cert.public_key().curve.name
        VERIFY_OR_RAISE(subject_public_key_info == "secp256r1",
                       f"The algorithm in subjectPublicKeyInfo field must be prime256v1, got {subject_public_key_info}")

        try:
            basic_constraints = pai_cert.extensions.get_extension_for_oid(x509.ExtensionOID.BASIC_CONSTRAINTS)
            VERIFY_OR_RAISE(basic_constraints.critical, "PAI Basic Constraints extension must be marked as critical")
            VERIFY_OR_RAISE(basic_constraints.value.ca, "PAI must be a CA certificate")
            VERIFY_OR_RAISE(basic_constraints.value.path_length == 0,
                           f"PAI path length must be 0, got {basic_constraints.value.path_length}")

            key_usage = pai_cert.extensions.get_extension_for_oid(x509.ExtensionOID.KEY_USAGE)
            VERIFY_OR_RAISE(key_usage.critical, "PAI Key Usage extension must be marked as critical")
            VERIFY_OR_RAISE(key_usage.value.key_cert_sign, "PAI must have keyCertSign usage enabled")
            VERIFY_OR_RAISE(key_usage.value.crl_sign, "PAI must have cRLSign usage enabled")
            VERIFY_OR_RAISE(not (key_usage.value.content_commitment or
                                 key_usage.value.key_agreement or
                                 key_usage.value.key_encipherment or
                                 key_usage.value.data_encipherment), "PAI must not have unsupported key usage extensions")

            subject_key_id = pai_cert.extensions.get_extension_for_oid(x509.ExtensionOID.SUBJECT_KEY_IDENTIFIER)
            authority_key_id = pai_cert.extensions.get_extension_for_oid(x509.ExtensionOID.AUTHORITY_KEY_IDENTIFIER)
            VERIFY_OR_RAISE(subject_key_id, "PAI missing Subject Key Identifier extension")
            VERIFY_OR_RAISE(authority_key_id, "PAI missing Authority Key Identifier extension")
        except x509.ExtensionNotFound as e:
            logging.error(f"PAI missing required extension: {str(e)}")
            return False
        return True
    except Exception as e:
        logging.error(f"PAI certificate validation failed: {str(e)}")
        return False

def validate_paa_cert(paa_cert_file: str) -> bool:
    """
    Validate the PAA certificate.
    Args:
        paa_cert_file (str): Path to the PAA certificate file.
    Returns:
        bool: True if the PAA certificate is valid, False otherwise.
    """
    try:
        paa_cert = load_cert_from_file(paa_cert_file)

        VERIFY_OR_RAISE(paa_cert.version == CERTIFICATE_VERSION,
                       f"PAA certificate version must be {CERTIFICATE_VERSION}, got {paa_cert.version}")
        VERIFY_OR_RAISE(paa_cert.signature_algorithm_oid._name == SIGNATURE_ALGORITHM,
                       f"PAA certificate signature algorithm must be {SIGNATURE_ALGORITHM}, got {paa_cert.signature_algorithm_oid._name}")

        # Self-signed check
        VERIFY_OR_RAISE(paa_cert.subject == paa_cert.issuer, "PAA must be self-signed")

        pid = extract_pid(paa_cert.subject)
        VERIFY_OR_RAISE(pid is None, "PAA must not include a Product ID (PID)")

        subject_public_key_info = paa_cert.public_key().curve.name
        VERIFY_OR_RAISE(subject_public_key_info == "secp256r1",
                       f"The algorithm in subjectPublicKeyInfo field must be prime256v1, got {subject_public_key_info}")

        try:
            basic_constraints = paa_cert.extensions.get_extension_for_oid(x509.ExtensionOID.BASIC_CONSTRAINTS)
            VERIFY_OR_RAISE(basic_constraints.critical, "PAA Basic Constraints extension must be marked as critical")
            VERIFY_OR_RAISE(basic_constraints.value.ca, "PAA must be a CA certificate")
            if basic_constraints.value.path_length is not None:
                VERIFY_OR_RAISE(basic_constraints.value.path_length == 1,
                               f"PAA path length must be 1, got {basic_constraints.value.path_length}")

            key_usage = paa_cert.extensions.get_extension_for_oid(x509.ExtensionOID.KEY_USAGE)
            VERIFY_OR_RAISE(key_usage.critical, "PAA Key Usage extension must be marked as critical")
            VERIFY_OR_RAISE(key_usage.value.key_cert_sign, "PAA must have keyCertSign usage enabled")
            VERIFY_OR_RAISE(key_usage.value.crl_sign, "PAA must have cRLSign usage enabled")
            VERIFY_OR_RAISE(not (key_usage.value.content_commitment or
                                 key_usage.value.key_agreement or
                                 key_usage.value.key_encipherment or
                                 key_usage.value.data_encipherment), "PAA must not have unsupported key usage extensions")

            subject_key_id = paa_cert.extensions.get_extension_for_oid(x509.ExtensionOID.SUBJECT_KEY_IDENTIFIER)
            VERIFY_OR_RAISE(subject_key_id, "PAA missing Subject Key Identifier extension")
        except x509.ExtensionNotFound:
            logging.error("PAA missing extension")
            return False
        return True

    except Exception as e:
        logging.error(f"PAA certificate validation failed: {str(e)}")
        return False

def validate_certificate_chain(cert_file: str, issuer_cert_file: str) -> bool:
    """
    Validate the certificate chain.
    Args:
        cert_file (str): Path to the certificate file.
        issuer_cert_file (str): Path to the issuer certificate file.
    Returns:
        bool: True if the certificate chain is valid, False otherwise.
    """
    try:
        cert = load_cert_from_file(cert_file)
        issuer_cert = load_cert_from_file(issuer_cert_file)

        # Check issuer-subject chaining
        VERIFY_OR_RAISE(cert.issuer == issuer_cert.subject,
                       f"Issuer-subject mismatch: {cert.issuer} != {issuer_cert.subject}")

        # Check certificate validity periods
        VERIFY_OR_RAISE(issuer_cert.not_valid_before <= cert.not_valid_before <= issuer_cert.not_valid_after,
                       f"Certificate is not currently valid. Valid from {issuer_cert.not_valid_before} to {issuer_cert.not_valid_after}")
        VERIFY_OR_RAISE(issuer_cert.not_valid_before <= cert.not_valid_after <= issuer_cert.not_valid_after,
                       f"Certificate is not currently valid. Valid from {issuer_cert.not_valid_before} to {issuer_cert.not_valid_after}")

        # Check for authority key identifier and subject key identifier
        try:
            cert_authority_key_id = cert.extensions.get_extension_for_oid(x509.ExtensionOID.AUTHORITY_KEY_IDENTIFIER)
            issuer_cert_subject_key_id = issuer_cert.extensions.get_extension_for_oid(x509.ExtensionOID.SUBJECT_KEY_IDENTIFIER)
            VERIFY_OR_RAISE(cert_authority_key_id, "Certificate missing Authority Key Identifier extension")
            VERIFY_OR_RAISE(issuer_cert_subject_key_id, "Issuer certificate missing Subject Key Identifier extension")
            VERIFY_OR_RAISE(cert_authority_key_id.value.key_identifier == issuer_cert_subject_key_id.value.digest,
                           "Certificate Authority Key Identifier and Issuer certificate Subject Key Identifier must match")
        except x509.ExtensionNotFound as e:
            logging.error(f"Missing required key identifier extension: {str(e)}")
            return False

        # Verify certificate signature using issuer's public key
        try:
            issuer_cert.public_key().verify(
                cert.signature,
                cert.tbs_certificate_bytes,
                ec.ECDSA(cert.signature_hash_algorithm)
            )
        except Exception as e:
            logging.error(f"Signature verification failed: {str(e)}")
            return False

        try:
            issuer_constraints = issuer_cert.extensions.get_extension_for_oid(
                x509.ExtensionOID.BASIC_CONSTRAINTS
            )
            VERIFY_OR_RAISE(issuer_constraints.value.ca, "Issuer certificate is not a CA")
        except Exception as e:
            logging.error(f"Failed to check issuer basic constraints: {str(e)}")
            return False

        return True

    except Exception as e:
        logging.error(f"Certificate chain validation failed: {str(e)}")
        return False

def verify_certificate_private_key(cert_file: str, private_key_file: str) -> bool:
    """
    Validate the input certificate and private key provided through arguments.

    Args:
        cert_file (str): Path to the certificate file.
        private_key_file (str): Path to the private key file.
    Returns:
        bool: True if the private key matches the certificate, False otherwise.
    """
    try:
        cert = load_cert_from_file(cert_file)
        cert_public_key = cert.public_key()
        private_key = load_key_from_file(private_key_file)

        # Verify key types are compatible
        VERIFY_OR_RAISE(isinstance(cert_public_key, ec.EllipticCurvePublicKey), "Certificate public key is not an elliptic curve key")
        VERIFY_OR_RAISE(isinstance(private_key, ec.EllipticCurvePrivateKey), "Private key is not an elliptic curve key")

        # Verify key curves match
        VERIFY_OR_RAISE(cert_public_key.curve.name == private_key.curve.name,
                       f"Key curve mismatch: cert uses {cert_public_key.curve.name}, private key uses {private_key.curve.name}")

        # Test message to sign and verify
        test_message = b"Test message for key validation"

        # Sign with private key
        signature = private_key.sign(
            test_message,
            ec.ECDSA(hashes.SHA256())
        )

        # Verify with certificate's public key
        try:
            cert_public_key.verify(
                signature,
                test_message,
                ec.ECDSA(hashes.SHA256())
            )
            return True
        except InvalidSignature:
            logging.error("Signature verification failed - private key does not correspond to certificate")
            return False

    except Exception as e:
        logging.error(f"Failed to verify certificate private key: {str(e)}")
        return False

def validate_certificates(args):
    if args.pai:
        VERIFY_OR_EXIT(args.cert, "PAI certificate is required")
        VERIFY_OR_EXIT(validate_pai_cert(args.cert), "PAI certificate is not valid")
        # In case of dac certificate and dac privatekey as a input pai certificate private key can be optional
        if args.key:
            VERIFY_OR_EXIT(verify_certificate_private_key(args.cert, args.key), "PAI certificate and private key do not match")
        VERIFY_OR_EXIT(validate_vid_pid_with_pai_cert(args.vendor_id, args.product_id, args.cert), "PAI certificate VID and PID do not match input VID and PID")

    if args.paa:
        VERIFY_OR_EXIT(args.cert, "PAA certificate is required")
        VERIFY_OR_EXIT(validate_paa_cert(args.cert), "PAA certificate is not valid")
        VERIFY_OR_EXIT(args.key, "PAA private key is required")
        VERIFY_OR_EXIT(verify_certificate_private_key(args.cert, args.key), "PAA certificate and private key do not match")

    if args.dac_cert:
        VERIFY_OR_EXIT(args.dac_key, "DAC private key is required")
        VERIFY_OR_EXIT(args.cert, "PAI certificate is required")
        VERIFY_OR_EXIT(validate_dac_cert(args.dac_cert), "DAC certificate is not valid")
        VERIFY_OR_EXIT(verify_certificate_private_key(args.dac_cert, args.dac_key), "DAC certificate and private key do not match")
        VERIFY_OR_EXIT(validate_certificate_chain(args.dac_cert, args.cert), "DAC certificate chain is not valid")

    if (args.valid_from or args.lifetime) and args.cert:
        VERIFY_OR_EXIT(validate_certificate_validity(args.valid_from, args.lifetime, args.cert),
                      f"{'PAA' if args.paa else 'PAI'} Certificate validity period is outside the specified parameters (from: {args.valid_from}, lifetime: {args.lifetime} days)")
        if args.dac_cert:
            VERIFY_OR_EXIT(validate_certificate_validity(args.valid_from, args.lifetime, args.dac_cert),
                          f"DAC certificate validity period is outside the specified parameters (from: {args.valid_from}, lifetime: {args.lifetime} days)")



#!/usr/bin/env python3

# Copyright 2022 Espressif Systems (Shanghai) PTE LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains utilitiy functions for validating argument.
"""

import os
import re
import sys
import csv
import enum
import base64
import logging
import binascii
from bitarray import bitarray
from bitarray.util import ba2int

ROTATING_DEVICE_ID_UNIQUE_ID_LEN_BITS = 128
SERIAL_NUMBER_LEN = 16


# Lengths for manual pairing codes and qrcode
SHORT_MANUALCODE_LEN = 11
LONG_MANUALCODE_LEN = 21
QRCODE_LEN = 22


INVALID_PASSCODES = [00000000, 11111111, 22222222, 33333333, 44444444, 55555555,
                     66666666, 77777777, 88888888, 99999999, 12345678, 87654321]


class CalendarTypes(enum.Enum):
    Buddhist = 0
    Chinese = 1
    Coptic = 2
    Ethiopian = 3
    Gregorian = 4
    Hebrew = 5
    Indian = 6
    Islamic = 7
    Japanese = 8
    Korean = 9
    Persian = 10
    Taiwanese = 11

class ProductFinish(enum.Enum):
    other = 0
    matte = 1
    satin = 2
    polished = 3
    rugged = 4
    fabric = 5


class ProductColor(enum.Enum):
    black = 0
    navy = 1
    green = 2
    teal = 3
    maroon = 4
    purple = 5
    olive = 6
    gray = 7
    blue = 8
    lime = 9
    aqua = 10
    red = 11
    fuchsia = 12
    yellow = 13
    white = 14
    nickel = 15
    chrome = 16
    brass = 17
    copper = 18
    silver = 19
    gold = 20

def VERIFY_OR_EXIT(condition, message = None):
    """
    Verify a condition and exit if it fails.

    Args:
        condition: Boolean condition to verify
        message: Error message to log and include in exit
    """
    if not condition:
        if message:
            logging.error(message)
        sys.exit(1)

def VERIFY_OR_RAISE(condition, message=None):
    """
    Verify a condition and raise an exception if it fails.

    Args:
        condition: Boolean condition to verify
        message: Error message to log and include in exception

    Raises:
        AssertionError: If the condition is False
    """
    if not condition:
        if message:
            logging.error(message)
        raise AssertionError(message)

def vid_pid_str(vid, pid):
    return '_'.join([hex(vid)[2:], hex(pid)[2:]])


def disc_pin_str(discriminator, passcode):
    return '_'.join([hex(discriminator)[2:], hex(passcode)[2:]])


# Checks if the input string is a valid hex string
def ishex(s):
    try:
        int(s, 16)
        return True
    except ValueError:
        return False


# Validate the input string length against the min and max length
def check_str_range(s, min_len, max_len, name):
    VERIFY_OR_EXIT(not(s and ((len(s) < min_len) or (len(s) > max_len))), f'{name} must be between {min_len} and {max_len} characters')


# Validate the date format
def check_date_format(date_str):
    VERIFY_OR_EXIT(not(date_str and not re.match(r'^\d{8}?$', date_str[0:8])), "First 8 bytes should be in ISO 8601 format YYYYMMDD (e.g., 20250416), last 8 can be anything specific to the manufacturer")

# Validate the input integer range
def check_int_range(value, min_value, max_value, name):
    VERIFY_OR_EXIT(not(value and ((value < min_value) or (value > max_value))), f'{name} is out of range, should be in range [{min_value}, {max_value}]')



# Validates discriminator and passcode
def validate_commissionable_data(args):
    check_int_range(args.discriminator, 0x0000, 0x0FFF, 'Discriminator')
    check_int_range(args.discovery_mode, 0, 7, 'Discovery mode')
    if args.passcode is not None:
        VERIFY_OR_EXIT(not((args.passcode < 0x0000001 and args.passcode > 0x5F5E0FE) or (args.passcode in INVALID_PASSCODES)), f'Invalid passcode {args.passcode}')

# Validate the device instance information
def validate_device_instance_info(args):
    check_int_range(args.product_id, 0x0000, 0xFFFF, 'Product id')
    check_int_range(args.vendor_id, 0x0000, 0xFFFF, 'Vendor id')
    check_int_range(args.hw_ver, 0x0000, 0xFFFF, 'Hardware version')
    check_str_range(args.serial_num, 1, SERIAL_NUMBER_LEN, 'Serial number')
    check_str_range(args.vendor_name, 1, 32, 'Vendor name')
    check_str_range(args.product_name, 1, 32, 'Product name')
    check_str_range(args.hw_ver_str, 1, 64, 'Hardware version string')
    check_str_range(args.mfg_date, 8, 16, 'Manufacturing date')
    check_str_range(args.rd_id_uid, 32, 32, 'Rotating device Unique id')
    check_str_range(args.part_number, 1, 32, 'Part number')
    check_date_format(args.mfg_date)


# Validate the device information: calendar types and fixed labels
def validate_device_info(args):
    # Validate the input calendar types
    if args.calendar_types is not None:
        if not (set(args.calendar_types) <= set(CalendarTypes.__members__)):
            invalid_types = set(args.calendar_types).union(set(CalendarTypes.__members__)) - set(CalendarTypes.__members__)
            logging.error('Unknown calendar type/s: %s', invalid_types)
            logging.error('Supported calendar types: %s', ', '.join(CalendarTypes.__members__))
            sys.exit(1)

    if args.fixed_labels is not None:
        for fl in args.fixed_labels:
            # Validate fixed label format: <endpoint_id>/<label_name>/<label_value>
            # Examples of valid fixed labels:
            # "0/orientation/up"
            VERIFY_OR_EXIT(re.match(r'^([0-9a-fA-F]{1,4})/(.{1,16})/(.{1,16})$', fl), f'Invalid fixed label: {fl}')

# Validates the attestation related arguments
def validate_attestation_info(args):
    # DAC key and DAC cert both should be present or none
    VERIFY_OR_EXIT((args.dac_key is not None) == (args.dac_cert is not None), "dac_key and dac_cert should be both present or none")
    VERIFY_OR_EXIT(not(args.dac_key is not None and args.pai is False), "Please provide PAI certificate along with DAC certificate and DAC key")

    # Validate the input certificate type, if DAC is not present
    if args.dac_key is None and args.dac_cert is None:
        if args.paa:
            logging.info('Input Root certificate type PAA')
        elif args.pai:
            logging.info('Input Root certificate type PAI')
        else:
            logging.info('Do not include the device attestation certificates and keys in partition binaries')

        # Check if Key and certificate are present
        VERIFY_OR_EXIT(not((args.paa or args.pai) and (args.key is None or args.cert is None)), 'CA key and certificate are required to generate DAC key and certificate')


# Validates DS peripheral related arguments
def validate_ds_peripheral_info(args):
    VERIFY_OR_EXIT(not(args.ds_peripheral and args.target.lower() != "esp32h2"), "DS peripheral is only supported for esp32h2 target")
    VERIFY_OR_EXIT(not(args.ds_peripheral and args.efuse_key_id == -1), "--efuse-key-id <value> is required when -ds or --ds-peripheral option is used")
    VERIFY_OR_EXIT(args.ds_peripheral is None or (not args.port or args.count == 1),"Port not specified or number of partitions count is greater than 1")

# Validates few basic cluster related arguments: product-label and product-url
def validate_basic_cluster_info(args):
    check_str_range(args.product_label, 1, 64, 'Product Label')
    check_str_range(args.product_url, 1, 256, 'Product URL')

# Validates the input arguments, this calls the above functions
def validate_args(args):
    # csv and mcsv both should present or none
    VERIFY_OR_EXIT((args.csv is not None) == (args.mcsv is not None), "csv and mcsv should be both present or none")
    if args.mcsv is not None:
        # Read the number of lines in mcsv file
        with open(args.mcsv, 'r', newline='') as f:
            csv_reader = csv.reader(f)
            # Count rows properly even when fields contain newlines
            row_count = sum(1 for row in csv_reader)

        # Subtract 1 for the header row
        args.count = row_count - 1

    validate_commissionable_data(args)
    validate_device_instance_info(args)
    validate_device_info(args)
    validate_attestation_info(args)
    validate_ds_peripheral_info(args)
    validate_basic_cluster_info(args)

    # If discriminator/passcode/DAC/serial_number/rotating_device_id is present
    # then we are restricting the number of partitions to 1
    if (args.discriminator is not None
            or args.passcode is not None
            or args.dac_key is not None
            or args.serial_num is not None
            or args.rd_id_uid is not None):
        VERIFY_OR_EXIT(args.count == 1, 'Number of partitions should be 1 when discriminator or passcode or DAC or serial number or rotating device id is present')

    # either --dac-in-secure-cert is acceptable or all in secure cert
    if args.commissionable_data_in_secure_cert or args.rd_id_uid_in_secure_cert:
        all_in_secure_cert = args.dac_in_secure_cert and args.commissionable_data_in_secure_cert and args.rd_id_uid_in_secure_cert
        VERIFY_OR_EXIT(
            all_in_secure_cert,
            "If --commissionable-data-in-secure-cert or --rd-id-uid-in-secure-cert is provided, then --dac-in-secure-cert, --commissionable-data-in-secure-cert, and --rd-id-uid-in-secure-cert are mandatory",
        )

    logging.info('Number of manufacturing NVS images to generate: {}'.format(args.count))


# Supported Calendar types is stored as a bit array in one uint32_t.
def calendar_types_to_uint32(calendar_types):
    # In validate_device_info() we have already verified that the calendar types are valid
    result = bitarray(32, endian='little')
    result.setall(0)
    for calendar_type in calendar_types:
        result[CalendarTypes[calendar_type].value] = 1
    return ba2int(result)


# get_fixed_label_dict() converts the list of strings to per endpoint dictionaries.
# example input  : ['0/orientation/up', '1/orientation/down', '2/orientation/down']
# example outout : {'0': [{'orientation': 'up'}], '1': [{'orientation': 'down'}], '2': [{'orientation': 'down'}]}
def get_fixed_label_dict(fixed_labels):
    fl_dict = {}
    for fl in fixed_labels:
        _l = fl.split('/')

        VERIFY_OR_EXIT(len(_l) == 3, f'Invalid fixed label: {fl}')
        VERIFY_OR_EXIT(ishex(_l[0]), f'Invalid fixed label: {fl}')
        VERIFY_OR_EXIT((len(_l[1]) > 0 and len(_l[1]) < 16), f'Invalid fixed label: {fl}')
        VERIFY_OR_EXIT((len(_l[2]) > 0 and len(_l[2]) < 16), f'Invalid fixed label: {fl}')

        if _l[0] not in fl_dict.keys():
            fl_dict[_l[0]] = list()

        fl_dict[_l[0]].append({_l[1]: _l[2]})

    return fl_dict


# get_supported_modes_dict() converts the list of strings to per endpoint dictionaries.
# example with semantic tags
# input  : ['0/label1/1/"1\0x8000, 2\0x8000" 1/label2/1/"1\0x8000, 2\0x8000"']
# outout : {'1': [{'Label': 'label1', 'Mode': 0, 'Semantic_Tag': [{'value': 1, 'mfgCode': 32768}, {'value': 2, 'mfgCode': 32768}]}, {'Label': 'label2', 'Mode': 1, 'Semantic_Tag': [{'value': 1, 'mfgCode': 32768}, {'value': 2, 'mfgCode': 32768}]}]}

# example without semantic tags
# input  : ['0/label1/1 1/label2/1']
# outout : {'1': [{'Label': 'label1', 'Mode': 0, 'Semantic_Tag': []}, {'Label': 'label2', 'Mode': 1, 'Semantic_Tag': []}]}

def get_supported_modes_dict(supported_modes):
    output_dict = {}

    for mode_str in supported_modes:
        mode_label_strs = mode_str.split('/')
        mode = mode_label_strs[0]
        label = mode_label_strs[1]
        ep = mode_label_strs[2]

        semantic_tags = ''
        if (len(mode_label_strs) == 4):
            semantic_tag_strs = mode_label_strs[3].split(', ')
            semantic_tags = [{"value": int(v.split('\\')[0]), "mfgCode": int(v.split('\\')[1], 16)} for v in semantic_tag_strs]

        mode_dict = {"Label": label, "Mode": int(mode), "Semantic_Tag": semantic_tags}

        if ep in output_dict:
            output_dict[ep].append(mode_dict)
        else:
            output_dict[ep] = [mode_dict]

    return output_dict

def get_random_rd_id_uid_b64():
    return base64.b64encode(os.urandom(int(ROTATING_DEVICE_ID_UNIQUE_ID_LEN_BITS / 8))).decode('utf-8')

def get_random_rd_id_uid_hex_str():
    return binascii.b2a_hex(os.urandom(int(ROTATING_DEVICE_ID_UNIQUE_ID_LEN_BITS / 8))).decode('utf-8')

def b64_to_hex(b64_str):
    return binascii.hexlify(binascii.a2b_base64(b64_str)).decode()

def hex_to_b64(hex_str):
    return binascii.b2a_base64(bytes.fromhex(hex_str)).decode().strip()


#!/usr/bin/env python3

# Copyright 2025 Espressif Systems (Shanghai) PTE LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Helps generate the secure cert partition for Matter

Below is the data that may be stored in secure cert partition:
- Attestation data:
    - Device Attestation Certificate
    - Private Key corresponding to the Device Attestation Certificate
    - Product Attestation Intermediate Certificate
- Commissionable Data:
    - Discriminator
    - Iteration Count
    - Salt
    - Verifier
- Device Instance Information:
    - Unique Id for Rotating Device ID
"""

import sys
import io
import os
import enum
import click
import base64
import logging
import contextlib
from typing import Optional
from dataclasses import dataclass
from esp_secure_cert.tlv_format_construct import EspSecureCert
from esp_secure_cert.tlv_format import tlv_type_t

# Temporary workaround until esp-secure-cert-tool defines Matter-specific TLV values in tlv_format.py
MATTER_TLV_TYPE_1 = 201


@dataclass
class MatterSecureCert:
    dac_cert: bytes
    dac_private_key: bytes
    pai_cert: bytes
    ds_peripheral: bool = False
    efuse_key_id: Optional[int] = None
    discriminator: Optional[int] = None
    iteration_count: Optional[int] = None
    salt: Optional[str] = None
    verifier: Optional[str] = None
    rd_id_uid: Optional[str] = None

    def __post_init__(self):
        if not self.dac_cert or not self.dac_private_key or not self.pai_cert:
            raise ValueError("DAC cert, DAC private key, and PAI cert are required")

        if (self.ds_peripheral is False) != (self.efuse_key_id is None):
            raise ValueError("either both or none of ds-peripheral and efuse-key-id must be provided")

        # check if all the optional arguments are provided together
        required_args = [self.discriminator, self.iteration_count, self.salt, self.verifier, self.rd_id_uid]
        len_required_args = len(required_args)
        count_of_provided_args = sum(1 for v in required_args if v is not None)
        if count_of_provided_args != 0 and count_of_provided_args != len_required_args:
            raise ValueError("discriminator, iteration_count, salt, verifier, and rd_id_uid must be provided together")

        # set the flag to add optional entries
        self.add_optional_entries = count_of_provided_args == len_required_args

    class MatterTLVSubType(enum.IntEnum):
        DISCRIMINATOR = 0
        SPAKE2P_VERIFIER = 1
        SPAKE2P_SALT = 2
        SPAKE2P_ITERATION_COUNT = 3
        RD_ID_UID = 4
        # use 128 and onwards for random tlv subtypes
        RANDOM1 = 128
        RANDOM2 = 129

    @staticmethod
    def get_cert_entry_as_json(cert: bytes, is_pai: bool = False) -> dict:
        return {
            "tlv_type": tlv_type_t.ESP_SECURE_CERT_CA_CERT_TLV if is_pai else tlv_type_t.ESP_SECURE_CERT_DEV_CERT_TLV,
            "tlv_subtype": 0,
            "data_value": base64.b64encode(cert).decode("utf-8"),
            "data_type": "base64",
        }

    @staticmethod
    def get_private_key_entry_as_json(private_key: bytes, ds_peripheral: bool = False, efuse_key_id: Optional[int] = None) -> dict:
        entry = {
            "tlv_type": tlv_type_t.ESP_SECURE_CERT_PRIV_KEY_TLV,
            "tlv_subtype": 0,
            "data_value": base64.b64encode(private_key).decode("utf-8"),
            "data_type": "base64",
            "priv_key_type": "ecdsa_peripheral" if ds_peripheral else "plaintext",
            "ds_enabled": ds_peripheral,
            "key_size": 256,
            "efuse_key_file": None,  # should be auto-generated in lower layer
            "efuse_key": None,
            "efuse_id": efuse_key_id if efuse_key_id is not None else None,
            "algorithm": "ECDSA",
        }
        if efuse_key_id is not None:
            entry["efuse_key_id"] = efuse_key_id
        return entry

    @staticmethod
    def get_discriminator_entry_as_json(discriminator: int) -> dict:
        return {
            "tlv_type": MATTER_TLV_TYPE_1,
            "tlv_subtype": MatterSecureCert.MatterTLVSubType.DISCRIMINATOR,
            "data_value": base64.b64encode(discriminator.to_bytes(2, "little")).decode("utf-8"),
            "data_type": "base64",
        }

    @staticmethod
    def get_spake2p_verifier_entry_as_json(verifier_b64: str) -> dict:
        return {
            "tlv_type": MATTER_TLV_TYPE_1,
            "tlv_subtype": MatterSecureCert.MatterTLVSubType.SPAKE2P_VERIFIER,
            "data_value": verifier_b64,
            "data_type": "base64",
        }

    @staticmethod
    def get_spake2p_salt_entry_as_json(salt_b64: str) -> dict:
        return {
            "tlv_type": MATTER_TLV_TYPE_1,
            "tlv_subtype": MatterSecureCert.MatterTLVSubType.SPAKE2P_SALT,
            "data_value": salt_b64,
            "data_type": "base64",
        }

    @staticmethod
    def get_spake2p_iteration_count_entry_as_json(iteration_count: int) -> dict:
        return {
            "tlv_type": MATTER_TLV_TYPE_1,
            "tlv_subtype": MatterSecureCert.MatterTLVSubType.SPAKE2P_ITERATION_COUNT,
            "data_value": base64.b64encode(iteration_count.to_bytes(4, "little")).decode("utf-8"),
            "data_type": "base64",
        }

    @staticmethod
    def get_rd_id_uid_entry_as_json(rd_id_uid_b64: str) -> dict:
        return {
            "tlv_type": MATTER_TLV_TYPE_1,
            "tlv_subtype": MatterSecureCert.MatterTLVSubType.RD_ID_UID,
            "data_value": rd_id_uid_b64,
            "data_type": "base64",
        }

    @staticmethod
    def get_random_entry_as_json(tlv_subtype: int) -> dict:
        random_len = 32
        return {
            "tlv_type": MATTER_TLV_TYPE_1,
            "tlv_subtype": tlv_subtype,
            "data_value": base64.b64encode(os.urandom(random_len)).decode("utf-8"),
            "data_type": "base64",
        }

    @staticmethod
    def get_random1_entry_as_json() -> dict:
        return MatterSecureCert.get_random_entry_as_json(MatterSecureCert.MatterTLVSubType.RANDOM1)

    @staticmethod
    def get_random2_entry_as_json() -> dict:
        return MatterSecureCert.get_random_entry_as_json(MatterSecureCert.MatterTLVSubType.RANDOM2)

    def generate_partition(self, port: Optional[int] = None) -> Optional[str]:
        """
        Generate the secure cert partition for Matter
        """

        secure_cert = EspSecureCert()
        secure_cert.add_entry(MatterSecureCert.get_cert_entry_as_json(self.dac_cert, is_pai=False))
        secure_cert.add_entry(MatterSecureCert.get_cert_entry_as_json(self.pai_cert, is_pai=True))
        secure_cert.add_entry(
            MatterSecureCert.get_private_key_entry_as_json(self.dac_private_key, self.ds_peripheral, self.efuse_key_id)
        )

        if self.add_optional_entries:
            logging.debug("Adding all optional entries")
            secure_cert.add_entry(MatterSecureCert.get_discriminator_entry_as_json(self.discriminator))
            secure_cert.add_entry(MatterSecureCert.get_spake2p_verifier_entry_as_json(self.verifier))
            secure_cert.add_entry(MatterSecureCert.get_spake2p_salt_entry_as_json(self.salt))
            secure_cert.add_entry(MatterSecureCert.get_spake2p_iteration_count_entry_as_json(self.iteration_count))
            secure_cert.add_entry(MatterSecureCert.get_rd_id_uid_entry_as_json(self.rd_id_uid))
            # add random entries only when we are adding optional entries
            secure_cert.add_entry(MatterSecureCert.get_random1_entry_as_json())
            secure_cert.add_entry(MatterSecureCert.get_random2_entry_as_json())

        # at the moment we only support esp32h2 for ds-peripheral
        target_chip = "esp32h2" if self.ds_peripheral else None

        output_buf = io.StringIO()
        output_stream = sys.stderr if logging.getLogger().level <= logging.DEBUG else output_buf
        try:
            with contextlib.redirect_stdout(output_stream), contextlib.redirect_stderr(output_stream):
                return secure_cert.generate_esp_secure_cert(target_chip, port)
        except SystemExit as e:
            logging.error(output_buf.getvalue())
            raise RuntimeError(f"ERROR: esp-secure-cert-tool exited with error:{e.code}")


# here onwards, cli functionality to test this script in a standalone manner
# TODO: Once the tlv_parse.py lands, we can use that to parse the secure cert partition
# and write some unit tests for the same


class AnyBaseInt(click.ParamType):
    name = "int (any base)"

    def convert(self, value, param, ctx):
        try:
            return int(value, 0)
        except ValueError:
            self.fail(f"{value} is not a valid integer", param, ctx)


any_base_int = AnyBaseInt()


@click.command()
@click.option("--dac", type=click.File("rb"), required=True, help="Path to the DAC certificate")
@click.option("--dac-key", type=click.File("rb"), required=True, help="Path to the DAC private key")
@click.option("--pai", type=click.File("rb"), required=True, help="Path to the PAI certificate")
@click.option("--ds", is_flag=True, help="Whether the DAC private key is for DS peripheral")
@click.option("--efuse-key-id", type=int, help="EFUSE key id")
@click.option("--port", help="Port to flash the secure cert partition")
@click.option("--discriminator", type=any_base_int, help="Discriminator")
@click.option("--iterations", type=any_base_int, help="Iteration count")
@click.option("--salt", help="spake2p salt encoded as base64 string")
@click.option("--verifier", help="spake2p verifier encoded as base64 string")
@click.option("--rd-id-uid", help="unique id for rotating device id encoded as base64 string")
def main(dac, dac_key, pai, ds, efuse_key_id, port, discriminator, iterations, salt, verifier, rd_id_uid):
    dac = dac.read()
    dac_key = dac_key.read()
    pai = pai.read()
    obj = MatterSecureCert(dac, dac_key, pai, ds, efuse_key_id, discriminator, iterations, salt, verifier, rd_id_uid)
    secure_cert_bin_name = obj.generate_partition(port)
    print(f"Secure cert binary generated: {secure_cert_bin_name}")


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    main()
